<?php

declare(strict_types=1);

namespace Drupal\languagewire_translation_provider\Test\Cron\Job;

use Drupal\languagewire_translation_provider\Adapter\SystemInterface;
use Drupal\languagewire_translation_provider\Adapter\TMGMT\TmgmtTranslatorInterface;
use Drupal\languagewire_translation_provider\Client\ClientInterface;
use Drupal\languagewire_translation_provider\Client\LanguageWireClient;
use Drupal\languagewire_translation_provider\Client\LanguageWireClientFactory;
use Drupal\languagewire_translation_provider\Cron\Job\ContentValidationJob;
use Drupal\languagewire_translation_provider\Database\ProjectRepositoryInterface;
use Drupal\languagewire_translation_provider\Domain\ProjectStatus;
use Drupal\languagewire_translation_provider\Service\ProjectContentValidator;
use Drupal\languagewire_translation_provider\Test\Builder\Domain\ProjectBuilder;
use PHPUnit\Framework\TestCase;
use Prophecy\Argument;
use Psr\Log\LoggerInterface;

/**
 * Content Validation Job Test .
 */
class ContentValidationJobTest extends TestCase {
  use \Prophecy\PhpUnit\ProphecyTrait;

  /**
   * System .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $system;

  /**
   * Logger .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $logger;

  /**
   * Project Repository .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $projectRepository;

  /**
   * Project Content Validator .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $projectContentValidator;

  /**
   * Client Factory .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $clientFactory;

  /**
   * Set Up .
   */
  protected function setUp(): void {
    $this->system = $this->prophesize(SystemInterface::class);
    $this->logger = $this->prophesize(LoggerInterface::class);
    $this->projectRepository = $this->prophesize(ProjectRepositoryInterface::class);
    $this->projectContentValidator = $this->prophesize(ProjectContentValidator::class);
    $this->system->logger()->willReturn($this->logger->reveal());
    $this->clientFactory = $this->prophesize(LanguageWireClientFactory::class);
    $this->clientFactory->createForTmgmtTranslator(Argument::type(TmgmtTranslatorInterface::class))->willReturn($this->prophesize(LanguageWireClient::class)->reveal());
  }

  /**
   * When Project Validation Is Successful Then Project Is Saved .
   *
   * @test
   */
  public function whenProjectValidationIsSuccessfulThenProjectIsSaved(): void {
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->build();
    $this->projectRepository->getAllByStatus(ProjectStatus::validating())->willReturn([$project]);
    $this->projectContentValidator->validateProject($project, $project->getTmgmtJob(), Argument::type(ClientInterface::class), $this->logger->reveal())->shouldBeCalled();
    $sut = $this->contentValidationJob();
    $sut->execute();
    $this->projectRepository->save($project)->shouldHaveBeenCalled();
  }

  /**
   * When Project Validation Throws Then Exception Is Logged .
   *
   * @test
   * @dataProvider throwableDataProvider
   */
  public function whenProjectValidationThrowsThenExceptionIsLogged(\Throwable $exception): void {
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->build();
    $this->projectRepository->getAllByStatus(ProjectStatus::validating())->willReturn([$project]);
    $this->projectContentValidator->validateProject($project, $project->getTmgmtJob(), Argument::type(ClientInterface::class), $this->logger->reveal())->willThrow($exception);
    $sut = $this->contentValidationJob();
    $sut->execute();
    $this->projectRepository->save($project)->shouldNotHaveBeenCalled();
    $this->logger->error(Argument::type('string'))->shouldHaveBeenCalled();
  }

  /**
   * Content Validation Job .
   */
  private function contentValidationJob(): ContentValidationJob {
    return new ContentValidationJob($this->system->reveal(), $this->projectRepository->reveal(), $this->clientFactory->reveal(), $this->projectContentValidator->reveal());
  }

  /**
   * Throwable Data Provider .
   */
  public function throwableDataProvider(): array {
    return [[new \Exception()], [new \Error()]];
  }

}
