<?php

declare(strict_types=1);

namespace Drupal\languagewire_translation_provider\Test\Service;

use Drupal\languagewire_translation_provider\Adapter\TMGMT\TmgmtJobInterface;
use Drupal\languagewire_translation_provider\Client\ClientInterface;
use Drupal\languagewire_translation_provider\Client\RecoverableException;
use Drupal\languagewire_translation_provider\Domain\DocumentStatus;
use Drupal\languagewire_translation_provider\Domain\ProjectStatus;
use Drupal\languagewire_translation_provider\Platform\Document\InvalidDocument;
use Drupal\languagewire_translation_provider\Platform\Document\UploadedDocument;
use Drupal\languagewire_translation_provider\Platform\Document\ValidatingDocument;
use Drupal\languagewire_translation_provider\Platform\Document\ValidDocument;
use Drupal\languagewire_translation_provider\Service\ProjectContentValidator;
use Drupal\languagewire_translation_provider\Test\Builder\Domain\DocumentBuilder;
use Drupal\languagewire_translation_provider\Test\Builder\Domain\ProjectBuilder;
use PHPUnit\Framework\TestCase;
use Prophecy\Argument;
use Psr\Log\LoggerInterface;

/**
 * Project Content Validator Test .
 */
class ProjectContentValidatorTest extends TestCase {
  use \Prophecy\PhpUnit\ProphecyTrait;

  /**
   * Client .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $client;

  /**
   * Logger .
   *
   * @var \Prophecy\Prophecy\ObjectProphecy
   */
  private $logger;

  /**
   * Set Up .
   */
  protected function setUp(): void {
    $this->client = $this->prophesize(ClientInterface::class);
    $this->logger = $this->prophesize(LoggerInterface::class);
  }

  /**
   * When There Is Nothing To Validate .
   *
   * @test
   */
  public function whenThereIsNothingToValidateAndThereAreOnlyValidDocumentsThenProjectStatusIsChangedToValidated(): void {
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->build();
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->assertEquals(ProjectStatus::validated(), $project->getStatus());
  }

  /**
   * When There Is Nothing To Validate With Invalid Documents .
   *
   * @test
   */
  public function whenThereIsNothingToValidateAndThereAreInvalidDocumentsThenProjectStatusIsChangedToError(): void {
    $invalidDocument = (new DocumentBuilder())->withStatus(DocumentStatus::error())->build();
    $validDocument = (new DocumentBuilder())->withStatus(DocumentStatus::validatedContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($validDocument, $invalidDocument)->build();
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->assertEquals(ProjectStatus::error(), $project->getStatus());
  }

  /**
   * When Recoverable Error Is Thrown During Request .
   *
   * @test
   */
  public function whenRecoverableErrorIsThrownDuringRequestingValidationThenWarningIsLoggedAndNothingIsUpdated(): void {
    $uploadedDocument = (new DocumentBuilder())->withStatus(DocumentStatus::uploadedContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($uploadedDocument)->build();
    $expectedException = new RecoverableException();
    $this->client->requestValidation([new UploadedDocument($uploadedDocument, $uploadedDocument->getPlatformId())])->willThrow($expectedException);
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->logger->warning(Argument::type('string'))->shouldHaveBeenCalled();
    $this->assertEquals(ProjectStatus::validating(), $project->getStatus());
    $this->assertEquals(DocumentStatus::uploadedContent(), $uploadedDocument->getStatus());
  }

  /**
   * When Recoverable Error Is Thrown During Validation .
   *
   * @test
   */
  public function whenRecoverableErrorIsThrownDuringValidationStatusCheckThenWarningIsLoggedAndNothingIsUpdated(): void {
    $validatingDocument = (new DocumentBuilder())->withStatus(DocumentStatus::validatingContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($validatingDocument)->build();
    $expectedException = new RecoverableException();
    $this->client->requestValidation([])->willReturn([]);
    $this->client->pollValidationStatus([new ValidatingDocument($validatingDocument)])->willThrow($expectedException);
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->logger->warning(Argument::type('string'))->shouldHaveBeenCalled();
    $this->assertEquals(ProjectStatus::validating(), $project->getStatus());
    $this->assertEquals(DocumentStatus::validatingContent(), $validatingDocument->getStatus());
  }

  /**
   * When Non Recoverable Error Is Throw During Request .
   *
   * @test
   */
  public function whenNonRecoverableErrorIsThrowDuringRequestingValidationThenErrorIsLoggedAndProjectStatusIsSetToErrorAndTmgmtJobIsMarkedUnprocessed(): void {
    $tmgmtJob = $this->prophesize(TmgmtJobInterface::class);
    $uploadedDocument = (new DocumentBuilder())->withStatus(DocumentStatus::uploadedContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($uploadedDocument)->build();
    $expectedException = new \Exception();
    $this->client->requestValidation([new UploadedDocument($uploadedDocument, $uploadedDocument->getPlatformId())])->willThrow($expectedException);
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $tmgmtJob->reveal(), $this->client->reveal(), $this->logger->reveal());
    $this->logger->error(Argument::type('string'))->shouldHaveBeenCalled();
    $this->assertEquals(ProjectStatus::error(), $project->getStatus());
    $tmgmtJob->makeUnprocessed()->shouldHaveBeenCalled();
    $tmgmtJob->save()->shouldHaveBeenCalled();
  }

  /**
   * When Non Recoverable Error Is Throw During Validation .
   *
   * @test
   */
  public function whenNonRecoverableErrorIsThrowDuringValidationStatusCheckThenErrorIsLoggedAndProjectStatusIsSetToErrorAndTmgmtJobIsMarkedUnprocessed(): void {
    $tmgmtJob = $this->prophesize(TmgmtJobInterface::class);
    $validatingDocument = (new DocumentBuilder())->withStatus(DocumentStatus::validatingContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($validatingDocument)->build();
    $expectedException = new \Exception();
    $this->client->requestValidation([])->willReturn([]);
    $this->client->pollValidationStatus([new ValidatingDocument($validatingDocument)])->willThrow($expectedException);
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $tmgmtJob->reveal(), $this->client->reveal(), $this->logger->reveal());
    $this->logger->error(Argument::type('string'))->shouldHaveBeenCalled();
    $this->assertEquals(ProjectStatus::error(), $project->getStatus());
    $tmgmtJob->makeUnprocessed()->shouldHaveBeenCalled();
    $tmgmtJob->save()->shouldHaveBeenCalled();
  }

  /**
   * When Documents Validation Was Requested .
   *
   * @test
   */
  public function whenDocumentsValidationWasRequestedThenDocumentsHaveValidatingStatusAndProjectStatusDoesNotChange(): void {
    $uploadedDocument = (new DocumentBuilder())->withStatus(DocumentStatus::uploadedContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($uploadedDocument)->build();
    $this->client->requestValidation([new UploadedDocument($uploadedDocument, $uploadedDocument->getPlatformId())])->will(function () use ($uploadedDocument) {
      return [new ValidatingDocument($uploadedDocument)];
    });
    $this->client->pollValidationStatus([])->willReturn([]);
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->assertEquals(ProjectStatus::validating(), $project->getStatus());
    $documents = $project->getDocuments();
    $this->assertCount(1, $documents);
    $this->assertEquals(DocumentStatus::validatingContent(), $documents[0]->getStatus());
  }

  /**
   * When Documents Validation Results With Errors .
   *
   * @test
   */
  public function whenDocumentsValidationResultsWithErrorsThenDocumentsStatusIsErrorAndProjectStatusIsErrorAndTmgmtJobIsMarkedUnprocessedWithMessage(): void {
    $tmgmtJob = $this->prophesize(TmgmtJobInterface::class);
    $validatingDocument = (new DocumentBuilder())->withStatus(DocumentStatus::validatingContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($validatingDocument)->build();
    $this->client->requestValidation([])->willReturn([]);
    $this->client->pollValidationStatus([new ValidatingDocument($validatingDocument)])->will(function () use ($validatingDocument) {
      return [new InvalidDocument($validatingDocument, 'Invalid')];
    });
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $tmgmtJob->reveal(), $this->client->reveal(), $this->logger->reveal());
    $this->assertEquals(ProjectStatus::error(), $project->getStatus());
    $documents = $project->getDocuments();
    $this->assertCount(1, $documents);
    $this->assertEquals(DocumentStatus::error(), $documents[0]->getStatus());
    $tmgmtJob->addMessage(Argument::type('string'))->shouldHaveBeenCalled();
    $tmgmtJob->makeUnprocessed()->shouldHaveBeenCalled();
    $tmgmtJob->save()->shouldHaveBeenCalled();
  }

  /**
   * When All Documents Are Valid .
   *
   * @test
   */
  public function whenAllDocumentsAreValidThenDocumentsStatusIdValidatedAndProjectStatusIsValidated(): void {
    $validatingDocument = (new DocumentBuilder())->withStatus(DocumentStatus::validatingContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($validatingDocument)->build();
    $this->client->requestValidation([])->willReturn([]);
    $this->client->pollValidationStatus([new ValidatingDocument($validatingDocument)])->will(function () use ($validatingDocument) {
      return [new ValidDocument($validatingDocument)];
    });
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->assertEquals(ProjectStatus::validated(), $project->getStatus());
    $documents = $project->getDocuments();
    $this->assertCount(1, $documents);
    $this->assertEquals(DocumentStatus::validatedContent(), $documents[0]->getStatus());
  }

  /**
   * When Project Has Uploaded And Validating Documents .
   *
   * @test
   */
  public function whenProjectHasUploadedAndValidatingDocumentsThenProjectStatusDoesNotChangeAndUploadedDocumentsAreNowValidating(): void {
    $uploadedDocument = (new DocumentBuilder())->withStatus(DocumentStatus::uploadedContent())->build();
    $validatingDocument = (new DocumentBuilder())->withStatus(DocumentStatus::validatingContent())->build();
    $project = (new ProjectBuilder())->withStatus(ProjectStatus::validating())->withDocuments($uploadedDocument, $validatingDocument)->build();
    $this->client->requestValidation([new UploadedDocument($uploadedDocument, $uploadedDocument->getPlatformId())])->will(function () use ($uploadedDocument) {
      return [new ValidatingDocument($uploadedDocument)];
    });
    $this->client->pollValidationStatus([new ValidatingDocument($validatingDocument)])->will(function () use ($validatingDocument) {
      return [new ValidDocument($validatingDocument)];
    });
    $sut = $this->projectContentValidator();
    $sut->validateProject($project, $project->getTmgmtJob(), $this->client->reveal(), $this->logger->reveal());
    $this->assertEquals(ProjectStatus::validating(), $project->getStatus());
    $this->assertEquals(DocumentStatus::validatingContent(), $uploadedDocument->getStatus());
    $this->assertEquals(DocumentStatus::validatedContent(), $validatingDocument->getStatus());
  }

  /**
   * Project Content Validator .
   */
  private function projectContentValidator(): ProjectContentValidator {
    return new ProjectContentValidator();
  }

}
